from test.nodes.test_accumulate import Accumulate
from test.nodes.test_add_value import AddValue
from test.nodes.test_below import Below
from test.nodes.test_double import Double
from test.nodes.test_greet import Greet
from test.nodes.test_merge import Merge
from test.nodes.test_remainder import Remainder
from test.nodes.test_rename import Rename
from test.nodes.test_repeat import Repeat
from test.nodes.test_subtract import Subtract
from test.nodes.test_sum import Sum
